/**
 * BER-TLV data object implementation (ISO 7816 Annex D.1)
 *
 * (C) Copyright (C) 2015, Datecs Ltd.
 *
 * No part of this software may be used, stored, compiled, reproduced,
 * modified, transcribed, translated, transmitted, or transferred, in any form
 * or by any means whether electronic, mechanical, magnetic, optical,
 * or otherwise, without the express prior written permission.
 */

#ifndef _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE 1
#endif

#include <stdio.h>
#include <stdarg.h>

#include "bertlv.h"

// Bitmask that specify constructed data object
#define MASK_CONSTRUCTED_DATA_OBJECT 0x20

// Bitmask that specify sequent bytes in tag.
#define MASK_SUBSEQUENT_BYTES 0x1F

// Bitmask that specify tag has more bytes.
#define MASK_MORE_BYTES 0x80

// Return if error occurs.
#define RETURN_ON_ERROR(X) if ((X) == BERTLV_ERROR) return BERTLV_ERROR;

// ----------------------------------------------------------------------------
// Copy specified number of bytes from one location to another.
// ----------------------------------------------------------------------------
static void* _bertlv_memcpy(void *dst, const void *src, bertlv_u32 size) 
{
	bertlv_u8* dst8 = (bertlv_u8*)dst;
    bertlv_u8* src8 = (bertlv_u8*)src;

    while ((size--) > 0) {
		*dst8++ = *src8++;
	}
    
	return dst;
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Get if value is valid tag
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_is_tag_valid(bertlv_u32 tag)
{	
	if (tag > 0xFFFFFF) {		
		if (((tag & (MASK_MORE_BYTES)))) return BERTLV_ERROR;
		if (!(tag & (MASK_MORE_BYTES << 8))) return BERTLV_ERROR;
		if (!(tag & (MASK_MORE_BYTES << 16))) return BERTLV_ERROR;
		if (((tag >> 24) & MASK_SUBSEQUENT_BYTES) != MASK_SUBSEQUENT_BYTES) {
			return BERTLV_ERROR;
		}
		return 4;
	} else if (tag > 0xFFFF) {
		if (((tag & (MASK_MORE_BYTES)))) return BERTLV_ERROR;
		if (!(tag & (MASK_MORE_BYTES << 8))) return BERTLV_ERROR;		
		if (((tag >> 16) & MASK_SUBSEQUENT_BYTES) != MASK_SUBSEQUENT_BYTES) {
			return BERTLV_ERROR;
		}
		return 3;
	} else if (tag > 0xFF) {
		if (((tag & (MASK_MORE_BYTES)))) return BERTLV_ERROR;	
		if (((tag >> 8) & MASK_SUBSEQUENT_BYTES) != MASK_SUBSEQUENT_BYTES) {
			return BERTLV_ERROR;
		}
		return 2;
	} else {
		if (((tag & MASK_SUBSEQUENT_BYTES)) == MASK_SUBSEQUENT_BYTES) {
			return BERTLV_ERROR;
		}
		return 1;
	}
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Get tag type.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_get_tag_type(bertlv_u32 tag)
{
	while (tag & (~0xFF)) tag >>= 8;

	if (tag & MASK_CONSTRUCTED_DATA_OBJECT) {
		return BERTLV_TAG_TYPE_CONSTRUCTED;
	} else {
		return BERTLV_TAG_TYPE_PRIMITIVE;
	}
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Get tag class.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_get_tag_class(bertlv_u32 tag)
{	
    while (tag & (~0xFF)) tag >>= 8;
		
	return (tag >> 6) & 0x03; // Returns last 2 bits of tag;
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Encode TLV tag.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_encode_tag(bertlv_u8 *buffer, bertlv_u32 tag)
{
	if (tag & 0xFF000000) {
		if (buffer) {
			*buffer++ = (bertlv_u8)(tag >> 24);
			*buffer++ = (bertlv_u8)(tag >> 16);
			*buffer++ = (bertlv_u8)(tag >> 8);
			*buffer = (bertlv_u8)(tag);
		}
		return 4;
	} else if (tag & 0xFF0000) {
		if (buffer) {
			*buffer++ = (bertlv_u8)(tag >> 16);
			*buffer++ = (bertlv_u8)(tag >> 8);
			*buffer = (bertlv_u8)(tag);
		}
		return 3;
	} else if (tag & 0xFF00) {
		if (buffer) {			
			*buffer++ = (bertlv_u8)(tag >> 8);
			*buffer = (bertlv_u8)(tag);
		}
		return 2;
	} else {
		if (buffer) {			
			*buffer = (bertlv_u8)(tag);
		}
		return 1;
	}	
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Encode TLV length.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_encode_length(bertlv_u8 *buffer, bertlv_u32 length)
{
	if (length & 0xFF000000) {
		if (buffer) {
			*buffer++ = (bertlv_u8)0x84; // 10000100
			*buffer++ = (bertlv_u8)(length >> 24);
			*buffer++ = (bertlv_u8)(length >> 16);
			*buffer++ = (bertlv_u8)(length >> 8);
			*buffer = (bertlv_u8)(length);
		}
		return 5;
	} else if (length & 0xFF0000) {
		if (buffer) {
			*buffer++ = (bertlv_u8)0x83; // 10000011
			*buffer++ = (bertlv_u8)(length >> 16);
			*buffer++ = (bertlv_u8)(length >> 8);
			*buffer = (bertlv_u8)(length);
		}
		return 4;
	} else if (length & 0xFF00) {
		if (buffer) {
			*buffer++ = (bertlv_u8)0x82; // 10000010			
			*buffer++ = (bertlv_u8)(length >> 8);
			*buffer = (bertlv_u8)(length);
		}
		return 3;
	} else {
		if (length >= 0x80) {
			if (buffer) {
				*buffer++ = (bertlv_u8)0x81; // 10000001							
				*buffer = (bertlv_u8)(length);
			}
			return 2;
		} else {
			if (buffer) {				
				*buffer = (bertlv_u8)(length);
			}
			return 1;
		}
	}
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Encode TLV tag and length (withouth data).
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_encode_tag_length(bertlv_u8 *buffer, bertlv_u32 tag, 
                                    bertlv_u32 length)
{
	bertlv_u32 err;
	bertlv_u32 off;
	
	err = bertlv_encode_tag(buffer, tag);
	RETURN_ON_ERROR(err);

	off = err;
	if (buffer) {
		err = bertlv_encode_length(buffer + err, length);
	} else {
		err = bertlv_encode_length(NULL, length);
	}
	RETURN_ON_ERROR(err);

	return (off + err);
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Encode tag and data into TLV.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_encode(bertlv_u8 *buffer, bertlv_u32 tag, bertlv_u32 length,
                         const bertlv_u8 *value)
{
	bertlv_u32 err;
	
	err = bertlv_encode_tag_length(buffer, tag, length);
	RETURN_ON_ERROR(err);

	if (length > 0 && value == NULL) {
		return BERTLV_ERROR;
	}

	if (buffer) {		
		_bertlv_memcpy(buffer + err, value, length);
	}

	return (err + length);
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Encode integer with two digits per byte and left-padded with zeros
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_encode_num(bertlv_u8 *buffer, bertlv_u32 tag, 
							 bertlv_u32 value, bertlv_u8 decimals)
{
	bertlv_u8 tmp[6];
	bertlv_u8 len;
	bertlv_u8 n;

	if (decimals == 0 || decimals > 12) 
		return BERTLV_ERROR;
	
	len = (decimals + 1) >> 1;

	n = 0;
	while (n < decimals) {
		if (n & 1) {
			tmp[len - 1 - n / 2] += (bertlv_u8)((value % 10) << 4);			
		} else {
			tmp[len - 1 - n / 2] = (bertlv_u8)(value % 10);
		}
		value = value / 10;
		n++;
	}
	
	return bertlv_encode(buffer, tag, len, tmp);
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Decode tag from buffer.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_decode_tag(const bertlv_u8 *buffer, bertlv_u32 count, 
							 bertlv_u32 *tag)
{
	const bertlv_u8 *ptr;
	const bertlv_u8 *end;
	bertlv_u32 val;
	
	if (buffer == NULL || count == 0) {
		return BERTLV_ERROR;
	}

	ptr = buffer;
	end = buffer + count;	
	val = *ptr++;	
	if ((val & MASK_SUBSEQUENT_BYTES) == MASK_SUBSEQUENT_BYTES) {
		do {
			if (ptr == end) {
				return BERTLV_ERROR; // Insufficient buffer size.
			}
			val = (val << 8) + (*ptr++);
		} while (val & MASK_MORE_BYTES);
	}

	if (tag) {
		*tag = val;
	}

	return (bertlv_u32)(ptr - buffer);	
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Decode length from buffer.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_decode_length(const bertlv_u8 *buffer, bertlv_u32 count, 
								bertlv_u32 *length)
{
	const bertlv_u8 *ptr;
	const bertlv_u8 *end;
	bertlv_u32 val;
	bertlv_u32 n;
	
	if (buffer == NULL || count == 0) {
		return BERTLV_ERROR;
	}

	ptr = buffer;
	end = buffer + count;	
	val = *ptr++;	
	if (val & MASK_MORE_BYTES) {
		n = val & (~MASK_MORE_BYTES);
		
		val = 0;
		do {
			if (ptr == end) {
				return BERTLV_ERROR; // Insufficient buffer size.
			}
			val = (val << 8) + (*ptr++);
		} while (--n);
	}

	if (length) {
		*length= val;
	}

	return (bertlv_u32)(ptr - buffer);	
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Decode tag and length from buffer.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_decode_tag_length(const bertlv_u8 *buffer, bertlv_u32 count, 
									bertlv_u32 *tag, bertlv_u32 *length)
{
	const bertlv_u8 *ptr;
	const bertlv_u8 *end;
	bertlv_u32 err;
		
	ptr = buffer;
	end = buffer + count;
	err = bertlv_decode_tag(ptr, end - ptr, tag);
	RETURN_ON_ERROR(err);
	
	ptr+= err;
	err = bertlv_decode_length(ptr, end - ptr, length);
	RETURN_ON_ERROR(err);
	
	return (ptr + err - buffer);
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Decode TLV.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_decode(const bertlv_u8 *buffer, bertlv_u32 count, 
						 bertlv_u32 *tag, bertlv_u32 *length, 
						 const bertlv_u8 **value)
{	
	bertlv_u32 err;
	bertlv_u32 n;
		
	err = bertlv_decode_tag_length(buffer, count, tag, &n);
	RETURN_ON_ERROR(err);

	if (count < (err + n)) {
		return BERTLV_ERROR; // Issuficient buffer size.
	}

	if (length) {
		*length = n;
	}

	if (value) {
		*value = buffer + err;
	}

	return err + n;
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Find a TLV by given tag value.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_find(const bertlv_u8 *buffer, bertlv_u32 count, 
                       bertlv_u32 tag, const bertlv_u8 **ptr)
{		
    bertlv_u32 err;
	bertlv_u32 val;
	bertlv_u32 off;
	
	off = 0;
	do {
		err = bertlv_decode(buffer + off, count - off, &val, NULL, NULL);
		RETURN_ON_ERROR(err);

		if (val == tag) {
			if (ptr) {
				*ptr = buffer + off;
			}
			return err;
		}

		off += err;
	} while (off < count);
	
	return BERTLV_ERROR; // Tag not found.
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Find TLV in buffer and return its value as numeric.
// ----------------------------------------------------------------------------
bertlv_u32 bertlv_find_value_num(const bertlv_u8 *buffer, 
								   bertlv_u32 count,
								   bertlv_u32 tag, 									    
								   bertlv_u32 defValue)
{
	bertlv_u8 *ptr;
	bertlv_u32 len;
	bertlv_u32 n;

	len = bertlv_find(buffer, count, tag, &ptr);
	if (len > 0) {
		bertlv_decode(ptr, len, NULL, &len, &ptr);

		defValue = 0;

		for (n = 0; n < len; n++) {
			defValue = (defValue << 8) + ptr[n];
		}
	} 

	return defValue;
	
}
// ----------------------------------------------------------------------------